#include "pad.h"
#include "main.h"
#include <math.h>

/** Copyright (C) 2010 Maxime Gamboni. See COPYING for
    copying/warranty info. */

#define STEPS 8

Pad::Pad(unsigned int id) : Item(id) {
  for (int i=0;i<FIRECOUNT;i++)
    firecube[i] = 0;
  can_fire = false;
}
Pad::~Pad() {}

void Pad::init(Level *lvl,double x,double y,int direction) {
  this->lvl = lvl;
  struct pad_info pi;
  pi.id       = 0;
  pi.cls      = CLS_PAD;
  pi.ev       = EV_INIT;
  pi.x        = x;
  pi.y        = y;
  pi.direction= direction;
  process((item_info*)&pi,0);

  poll();
}

void Pad::process(struct item_info *data,size_t len) {
  struct pad_info *pi = (struct pad_info*)data;

  switch (pi->ev) {
  case EV_INIT:
    this->x = pi->x;
    this->y = pi->y;
    this->direction = pi->direction;
    
    this->dirx = (direction==RIGHTPAD)?-1:(direction==LEFTPAD)?+1:0;
    this->diry = (direction==TOPPAD)?-1:(direction==BOTTOMPAD)?+1:0;
    
    this->targ = dirx?y:x;

    this->angle = dirx?HALFPI:0;
    //  this->captured = 0;

    reset();

    prevx = x;
    prevy = y;
    prevangle = angle;

    break;
  case EV_MOVED:
    this->targ = pi->x;
    break;
  }
}

void Pad::poll() {
  struct pad_info pi;
  pi.cls      = CLS_PAD;
  pi.ev       = EV_INIT;
  pi.x        = x;
  pi.y        = y;
  pi.direction= direction;
  push((item_info*)&pi,sizeof(struct pad_info));
}

int *Pad::get_direction() {
  static int d[2];
  d[0] = dirx;
  d[1] = diry;
  return d;
}

void Pad::update(double t) {

  if (t==0) return;

  dt = t; // see check_collision()

  //  printf("\n");

  t *= 20;
  if (t>1) t=1;

  if (w<tw) {
    w += t/10.0f;
    if (w>tw) w=tw; // went too far
  } else if (w>tw) {
    w -= t/10.0f;
    if (w<tw) w=tw; // went too far
  }

  prevangle=angle;
  prevx=x;
  prevy=y;

  if (dirx) angle = HALFPI-angle; // to have angle at rest be zero for
				  // the calculations below

  angle *= 1-t/20; // gradually bring the angle back to zero

  /* If the pad is far from the target then the angle should
     intensify accordingly */
  double a = (dirx?dirx*(y-targ):diry*(x-targ))/5;
  if (fabs(angle)<fabs(a)) {
    angle = angle*t/2 + a*(1-t/2);
  }

  if (dirx) angle = HALFPI-angle;

  if (dirx)
    y = t*targ + (1-t)*y;
  else
    x = t*targ + (1-t)*x;

  // if (captured) {
  //   captured->move(x+0.3*dirx,y+0.3*diry);
  // }

  for (int i=0;i<FIRECOUNT;i++)
    if (firecube[i])
      if (firecube[i]->update(t,lvl) || firecube[i]->has_vanished()) {
	delete (firecube[i]);
	firecube[i] = 0;
      }
}

void Pad::move(double tx,double ty) {
  if (dirx)
    this->targ = ty;
  else
    this->targ = tx;

  struct pad_info pi;
  pi.cls      = CLS_PAD;
  pi.ev       = EV_MOVED;
  pi.x        = targ;
  push((item_info*)&pi,sizeof(struct pad_info));

  lookAt(x,y,false); // only make sure the pad is visible when it is moved
}

bool Pad::shift(double dtx,double dty) {
  lookAt(x,y,false); // only make sure the pad is visible when it is moved
  if (dirx) {
    this->targ += dty;
    if (this->targ < 0) this->targ = 0;
    else if (this->targ > lvl->get_height())
      this->targ = lvl->get_height();

    return (dtx*dirx>0.4);
  } else {
    this->targ += dtx;
    if (this->targ < 0) this->targ = 0;
    else if (this->targ > lvl->get_width())
      this->targ = lvl->get_width();
    return (dty*diry>0.4);
  }
}

void Pad::capture(Ball *b) {
  //  this->captured = b;
  b->move(x+0.3*dirx,y+0.3*diry);
  //  captured->set_speed(0,0);
}

void Pad::trigger() {
  tock->play();
}

void Pad::activate() {
  // if (captured) {
  //   double s = dirx+diry;
  //   captured->set_speed(s,s);
  //   captured = 0;
  // }
  if (can_fire && tw > 0.05f) {
    tw -= 0.01f; // he he he he he :)
    for (int i=0;i<FIRECOUNT;i++)
      if (firecube[i]==0) {
	firecube[i] = new Littlecube(0);
	firecube[i]->init(x+0.1f*dirx,y+0.1f*diry,0.3,true,false);
	firecube[i]->set_speed(dirx,diry);
        break;
      }
  }
}

void Pad::grow() {
  tw += 0.8f;
}

void Pad::fire() {
  can_fire = true;
}

void Pad::reset() {
  w = 1;
  tw = 1;
  can_fire = false;
}

/* Pass the coordinates of three points a, b and c. Store in d the
   distance of c to the a-b line (positive or negative depending on
   the side), and in l the position of c's projection on the a-b line
   relative to a and b (a has coordinate zero, b has coordinate 1).

   You'll get a division by zero if a and b are equal. */
void proj(double ax, double ay, double bx, double by, double cx, double cy, double *d, double *l) {
  double ab = (bx-ax)*(bx-ax) + (by-ay)*(by-ay);

  /* Scalar product ac · ab divided by ab's sqlength */
  *l = ((cx-ax)*(bx-ax) + (cy-ay)*(by-ay)) / ab;

  /* Scalar product ac times ab's normal vector pointing right if b is
     on top, again divided by ab's sqlength to get actual distance. */
  *d = ((cx-ax)*(by-ay) + (cy-ay)*(ax-bx)) / ab;

}

/* See similarly named methods in level.h */
double Pad::check_collision(double x,double y,double vx,double vy, double t, double *n, double *v) {

  /* The pad is now at (this->x;this->y <· this->angle), and was dt
     seconds ago at (this->prevx;this->prevy <·
     this->prevangle). Using interpolation we
     find the position and angle t seconds ago: */
  double px = this->x-(this->x-prevx)*t/dt,
    py = this->y-(this->y-prevy)*t/dt,
    pa = angle-(angle-prevangle)*t/dt;

  /* "Right" end of the pad t seconds ago */
  double bx=px+cos(pa)*w, by=py+sin(pa)*w;

  /* Position of the object relative to the pad t seconds ago */
  double pd;
  double pl;

  proj(px,py,bx,by, x,y, &pd,&pl);

  // printf("(%f;0<%f)=distance %f projection %f ",px,pa,pd,pl);

  /* Right end of the pad in the current frame */
  bx=this->x+cos(angle)*w; by=this->y+sin(angle)*w;

  /* Position of the object relative to the pad in the current
     frame */
  double d;
  double l;
  proj(this->x,this->y,bx,by, x+vx*t,y+vy*t, &d,&l);

  // printf("-> o(%f;%f) -> (%f;0<%f)=distance %f projection %f",x,y,this->x,angle,d,l);

  /* The ball was on the same side of the pad before and after? No
     collision */
  if (pd*d > 0) return 2;

  // printf("_sign-change");

  /* Crossed the line of the pad, but didn't touch it */
  if (fabs(pl)>1 || fabs(l)>1) return 2;

  // printf("_collision");

  /* If we reach this point, there *was* a collision. But where? */

  /* To simplify we assume that the distance of the object to the pad
     decreased at a constant rate until reaching zero at collision
     time, then increased at the same constant rate until reaching the
     final position. Therefore the collision occurred at this time: */
  double ret = pd/(pd-d); // between zero (pd=0) and one (d=0)

  /* Interpolate the angle to get the normal vector at collision
     time */
  double collang = pa + (angle-pa)*ret;
  n[0] = -sin(collang);
  n[1] = cos(collang);

  // printf("_n=(%f;%f)",n[0],n[1]);

  /* To get the speed at collision time we need to estimate which part
     of the pad hit the object. Interpolating: */
  double collpos = pl + (l-pl)*ret;

  /* First term: intrinsic speed of the pad. Second component:
  Angular speed at that position.

  Note that using prev and dt instead of p and t should give the same
  values. */
  v[0]=((this->x-px) + n[0]*collpos*w*(angle-pa))/t;
  v[1]=((this->y-py) + n[1]*collpos*w*(angle-pa))/t;

  /* positive pd means the object was on the "bottom" of the pad
     before the collision, and on its "top" after it, so the normal
     needs to point "down" */
  if (pd>0) {// printf("_reversal");
    n[0] = -n[0]; n[1] = -n[1];
  }

  // printf("rc=%f",t);

  return ret;

  /*** OLD CODE FOLLOWS ***/
//   double shortest=2;
//   double t;
 
// /* So we actually run the same code twice once with the actual
//    variables and once with the prev- variables */

//   for (int pass=0;pass<2;pass++) {

//     /* what n will be set to if there's a collision */
//     double nx = -sin(angle)*dirsign;
//     double ny = cos(angle)*dirsign;
    
//     /* distance of the source to the pad line */
//     double dist = (x-this->x) * nx +
//       (y-this->y) * ny;
    
//     if (dist>=0) {
//       /* the ball is above the pad line */
    
//       /* What we'll return if there's a collision */
//       double d = - dist / (dx*nx + dy*ny);

//       if (d>= 0 && d<=1 && d < shortest) {
  
// 	/* if the squared distance of the potential collision point from the
// 	   centre of the pad is less than w squared then there is a
// 	   collision. */
	
// 	if ( sq(x+dx*d-this->x) + 
// 	     sq(y+dy*d-this->y) < w*w ) {
// 	  n[0] = nx;
// 	  n[1] = ny;
// 	  shortest = d;
// 	}
//       }
//     }

//     t=prevx;prevx=this->x;this->x=t;
//     t=prevy;prevy=this->y;this->y=t;
//     t=prevangle;prevangle=angle;angle=t;
//   }

//   return shortest;
}

void Pad::render() {
  double redbits = w>0.31? 0.3 : w-0.01;
  glPushMatrix(); {
    glTranslatef(x,y,0);
    glRotatef(angle RADIANS,0,0,1);
    
    //    glEnable(GL_LIGHTING);


    if (redbits <= 0.1) {
      /* The entire pad is a red cube if it's too small */
      glColor3f(1,0,0);
      glTranslatef(-w,-0.1,-0.5);
      render_cube(2*w,0.2,1,true);
    } else {
      glColor3f(1,1,1);

      glTranslatef(-w+0.05,-0.1,-0.5);
      render_cube(2*w-0.1,0.2,1,true);

      glColor3f(1,0,0);
      glTranslatef(-0.05,-0.05,-0.05);
      render_cube(redbits,0.3,1.1,true);

      glTranslatef(2*w-redbits,0,0);
      render_cube(redbits,0.3,1.1,true);
    }
    // glNormal3f( 0,0, 1); // front
    // glVertex3f(-w,-0.1f, 0.5f);
    // glVertex3f(-w,0.1f, 0.5f);
    // glVertex3f( w,0.1f, 0.5f);
    // glVertex3f( w,-0.1f, 0.5f);

    // glNormal3f( 1,0, 0); // right
    // glVertex3f( w,-0.1f, 0.5f);
    // glVertex3f( w,0.1f, 0.5f);
    // glVertex3f( w,0.1f,-0.5f);
    // glVertex3f( w,-0.1f,-0.5f);

    // glNormal3f( 0,0,-1); // back (is this even ever shown?)
    // glVertex3f( w,-0.1f,-0.5f);
    // glVertex3f( w,0.1f,-0.5f);
    // glVertex3f(-w,0.1f,-0.5f);
    // glVertex3f(-w,-0.1f,-0.5f);

    // glNormal3f(-1,0, 0); // left
    // glVertex3f(-w,-0.1f,-0.5f);
    // glVertex3f(-w,0.1f,-0.5f);
    // glVertex3f(-w,0.1f, 0.5f);
    // glVertex3f(-w,-0.1f, 0.5f);

    // glNormal3f( 0,1,0); // top
    // glVertex3f(-w,0.1f,-0.5f);
    // glVertex3f( w,0.1f,-0.5f);
    // glVertex3f( w,0.1f,0.5f);
    // glVertex3f(-w,0.1f,0.5f);

    glNormal3f( 0,-1,0); // bottom
    glVertex3f(-w,-0.1f,-0.5f);
    glVertex3f( w,-0.1f,-0.5f);
    glVertex3f( w,-0.1f, 0.5f);
    glVertex3f(-w,-0.1f, 0.5f);

  } glPopMatrix();

  glColor3f(1,0,0);
  for (int i=0;i<FIRECOUNT;i++)
    if (firecube[i])
      firecube[i]->render();

  //  glDisable(GL_LIGHTING);

}
