#include "server.h"

/** Copyright (C) 2010 Maxime Gamboni. See COPYING for
    copying/warranty info. */

#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>

#include "item.h"
#include "main.h"

Server::Server() {
  sfd = -1;
  cfd = -1;
}
Server::~Server() {
  if (cfd != -1) close(cfd);
  if (sfd != -1) close(sfd);
}

bool Server::server() {
  struct sockaddr_in my_addr;
  sfd = socket(AF_INET, SOCK_STREAM, 6); // 6==TCP
  if (sfd==-1) { printf("Failed creating socket (%s)\n",strerror(errno));return false; }

  if (fcntl(sfd,F_SETFL,O_NONBLOCK)==-1) { printf("Warning: will block until client connects (%s).\n",strerror(errno)); }

  memset(&my_addr, 0, sizeof(struct sockaddr_in));
  my_addr.sin_family = AF_INET;
  my_addr.sin_port = htons(PORT);
  my_addr.sin_addr.s_addr = INADDR_ANY;
  if (bind(sfd,(struct sockaddr *) &my_addr,sizeof(struct sockaddr_in))) {
    printf("Failed binding socket (%s)\n",strerror(errno)); return false;
  }
  if (listen(sfd,PEERS)) {
    printf("Failed opening socket for listening (%s)\n",strerror(errno)); return false;
  }

  printf("Server waiting for connection on port %d.\n",PORT);
  return true;
}

bool Server::client(const char *s) {
  bool success=true;
  struct addrinfo *your_addr;
  cfd = socket(AF_INET, SOCK_STREAM, 6); // 6==TCP
  if (cfd==-1) {
    printf("Failed creating socket (%s)\n",strerror(errno));
    success=false;
  }

  int r = getaddrinfo(s,SPORT,0,&your_addr);
  if (r!=0) {
    printf("Couldn't resolve %s (%s).\n",s,gai_strerror(r));
    success=false;
  } else {
    r = connect(cfd,your_addr->ai_addr,your_addr->ai_addrlen);
    if (r!=0) {
      printf("%s: connection failed (%s).\n",s,gai_strerror(r));
      success=false;
    }
  }
  freeaddrinfo(your_addr);

  if (fcntl(cfd,F_SETFL,O_NONBLOCK)==-1) {
    printf("Warning: couldn't make cfd non-blocking (%s).\n",strerror(errno));
  }

  return success;
}

void Server::push(struct item_info *data,size_t len) {
  if (cfd != -1) {
    unsigned int llen = len+sizeof(unsigned int); // data length including the length header
    char *buf = (char*)malloc(llen);
    unsigned int nlen = htonl(len); // network byte ordered len

    memcpy(buf,&nlen,sizeof(unsigned int));
    memcpy(buf+sizeof(unsigned int),data,len);
    size_t r = send(cfd,buf,llen,0);
    if (r>0) {
      if (llen != r)
	printf("Having %d bytes only sent %d...\n",llen,r);
    }
    free(buf);
  }
}

void Server::poll() {
  struct sockaddr_in your_addr;
  socklen_t addrlen = sizeof(struct sockaddr_in);
  if (cfd==-1) {
    /* No client yet. See if one is waiting */
    cfd = accept(sfd,(struct sockaddr *) &your_addr,&addrlen);
    if (cfd==-1) {
      return; // still no one
    } else {
      if (fcntl(cfd,F_SETFL,O_NONBLOCK)==-1) {
	printf("Warning: couldn't make cfd non-blocking (%s).\n",strerror(errno));
      }
      printf("Client connected from %s.\n", inet_ntoa(your_addr.sin_addr));

      level->poll();
    }
    /* Note, data from the client will only be read() on next call to poll() */
  } else {
    //    bool g=false; // debug - whether we got at least a packet
    for (;;) { /* Exiting when there's no more data */
      int len, r;
      r = recv(cfd,&len,sizeof(unsigned int),0);
      if (r>0) {
	len = ntohl(len);

	//	g=true;
	//	printf("Receiving %d + %d bytes...\n",r,len);
	if (len>0) {
	  item_info *data = (item_info*)malloc(len);
	  r = recv(cfd,data,len,0);
	  if (r==len) {
	    Item *it = level->get_item(data->id);
	    if (it)
	      it->process(data,len);
	    else
	      new_Item(data,len);

	  } else printf("Expected %d bytes only got %d...\n",len,r);

	  free(data);
	} else {
	  printf("Got zero-sized packet!\n");
	  return;
	}
      } else {
	//	if (g) printf("\n");
	return;
      }
    }
  }
}
